/* Copyright (C) 2005-2019 Shugo Maeda <shugo@ruby-lang.org> and Charlie Savage
   <cfis@savagexi.com> Please see the LICENSE file for copyright and
   distribution information */

#include "bf_method.h"
#include "bf_allocation.h"
#include "bf_call_trees.h"
#include "bf_profile.h"

#define SINGLETON_CLASS_METHOD_FLAG 2 // <Class:MyObject>#test 
#define SINGLETON_MODULE_METHOD_FLAG 4 // <Module:MyObject>#test
#define SINGLETON_OBJ_METHOD_FLAG 8 // <Object:MyObject>#test

const size_t MAX_METHOD_LEN = 1024;

VALUE cRpMethodInfo;

/* ================  Helper Functions  =================*/
VALUE resolve_klass(VALUE klass, unsigned int *klass_flags)
{
    VALUE result = klass;

    if (klass == 0 || klass == Qnil)
    {
        result = Qnil;
    }
    else if (BUILTIN_TYPE(klass) == T_CLASS && FL_TEST(klass, FL_SINGLETON))
    {
        /* We have come across a singleton object. First
           figure out what it is attached to.*/
        VALUE attached = rb_iv_get(klass, "__attached__");

        /* Is this a singleton class acting as a metaclass? */
        if (BUILTIN_TYPE(attached) == T_CLASS)
        {
            *klass_flags |= kClassSingleton;
            result = attached;
        }
        /* Is this for singleton methods on a module? */
        else if (BUILTIN_TYPE(attached) == T_MODULE)
        {
            *klass_flags |= kModuleSingleton;
            result = attached;
        }
        /* Is this for singleton methods on an object? */
        else if (BUILTIN_TYPE(attached) == T_OBJECT)
        {
            *klass_flags |= kObjectSingleton;
            result = rb_class_superclass(klass);
        }
        /* Ok, this could be other things like an array made put onto
           a singleton object (yeah, it happens, see the singleton
           objects test case). */
        else
        {
            *klass_flags |= kOtherSingleton;
            result = klass;
        }
    }
    /* Is this an include for a module?  If so get the actual
       module class since we want to combine all profiling
       results for that module. */
    else if (BUILTIN_TYPE(klass) == T_ICLASS)
    {
        unsigned int dummy;
        *klass_flags |= kModuleIncludee;
        result = resolve_klass(RBASIC(klass)->klass, &dummy);
    }
    return result;
}

VALUE resolve_klass_name(VALUE klass, const unsigned int *klass_flags)
{
    VALUE result = Qnil;

    if (klass == Qnil)
    {
        result = rb_str_new2("[global]");
    }
    else if (*klass_flags & kOtherSingleton)
    {
        result = rb_any_to_s(klass);
    }
    else
    {
        result = rb_class_name(klass);
    }

    return result;
}

// gets fnargs for the function(if defined) and return the arguments as URL 
// encoded string
static VALUE get_fnargs(prof_profile_t *profile, VALUE meth_name)
{
    VALUE fnarg_ids;
    const size_t MAX_EVAL_BLOCK_SIZE = 255;
    char eval_block[MAX_EVAL_BLOCK_SIZE];
    char arg_val_copy[MAX_METHOD_LEN];
    char *arg_name;
    const char *fnarg_prefix = "";
    VALUE arg_val, result;

    fnarg_ids = rb_hash_aref(profile->fnarg_rules, meth_name);
    if (fnarg_ids == Qnil || !RB_TYPE_P(fnarg_ids, T_ARRAY)) {
        return Qnil;
    }

    result = rb_str_new2("");

    // Note: Use this to debug print the local variables/func args
    //rb_eval_string("puts 'locals=', local_variables");

    for(long i=0; i < RARRAY_LEN(fnarg_ids); i++) {
        VALUE v = rb_ary_entry(fnarg_ids, i);

        if RB_TYPE_P(v, T_FIXNUM) { // positional arg
            snprintf(eval_block,
                    MAX_EVAL_BLOCK_SIZE,
                    "eval(local_variables[%u].to_s)",
                    NUM2UINT(v) - 1); // fn-args starts from 1
            arg_val = rb_eval_string(eval_block);
        } else { // named arg
            arg_name = StringValuePtr(v);
            snprintf(eval_block,
                    MAX_EVAL_BLOCK_SIZE,
                    "if local_variables.include? '%s'.to_sym\n"
                    "    eval '%s'\n"
                    "end",
                    arg_name, arg_name);
            arg_val = rb_eval_string(eval_block);
        }

        // Note: Use this to debug print the code executed by rb_eval
        //printf("eval block:\n%s\n", eval_block);

        // arg index/name might be invalid continue in that situation
        if (arg_val == Qnil) {
            continue;
        }

        // replace bad chars from the fn_args
        arg_val = rb_sprintf("%"PRIsVALUE, arg_val); // convert the arg_val to a string
        char *arg_val_cstr = StringValuePtr(arg_val);
        memset(arg_val_copy, 0, MAX_METHOD_LEN);
        size_t dst_i = 0;
        for (int j=0; j < RSTRING_LEN(arg_val) && dst_i < MAX_METHOD_LEN; j++)
        {
            switch (arg_val_cstr[j]) {
            case '%':
                memcpy(&arg_val_copy[dst_i], "%25", 3);
                break;
            case '+':
                memcpy(&arg_val_copy[dst_i], "%2B", 3);
                break;
            case '=':
                memcpy(&arg_val_copy[dst_i], "%3D", 3);
                break;
            case '@':
                memcpy(&arg_val_copy[dst_i], "%40", 3);
                break;
            case '&':
                memcpy(&arg_val_copy[dst_i], "%26", 3);
                break;
            case '\r':
                memcpy(&arg_val_copy[dst_i], "%0D", 3);
                break;
            case '\n':
                memcpy(&arg_val_copy[dst_i], "%0A", 3);
                break;
            default:
                arg_val_copy[dst_i] = arg_val_cstr[j];
                dst_i += 1;
                continue;
            }
            dst_i += 3;
        }

        // all fn_args will prepend an ampersand except the first one
        if (i > 0) {
            fnarg_prefix = "&";
        } else {
            fnarg_prefix = "?";
        }

        result = rb_sprintf("%"PRIsVALUE"%s%"PRIsVALUE"=%s", result, fnarg_prefix, 
                            v, arg_val_copy);
    }
    return result;
}

st_data_t method_key(VALUE klass, VALUE msym, void *profile_ptr, method_key_t *mkey)
{
    VALUE resolved_klass = klass;
    unsigned int klass_flags = 0;
    char *result = ALLOC_N(char, MAX_METHOD_LEN);
    const char *mname, *cname;
    prof_profile_t *profile = (prof_profile_t *)profile_ptr;

    mkey->fqmn = result;

    // We needed to resolve the klass_name here instead of method create because
    // then it will hash into same slot functions like Class#new for every class.
    resolved_klass = resolve_klass(klass, &klass_flags);

    mname = (msym != Qnil ? rb_id2name(SYM2ID(msym)) : "");
    cname = (resolved_klass != Qnil ? rb_class2name(resolved_klass) : "");

    // format accordinly if the method is singleton class/module/obj
    mkey->klass = result;
    switch (klass_flags) {
        case SINGLETON_CLASS_METHOD_FLAG:
            mkey->klass_len = snprintf(result, MAX_METHOD_LEN, "<Class::%s>", cname);
            break;
        case SINGLETON_MODULE_METHOD_FLAG:
            mkey->klass_len = snprintf(result, MAX_METHOD_LEN, "<Module::%s>", cname);
            break;
        case SINGLETON_OBJ_METHOD_FLAG:
            mkey->klass_len = snprintf(result, MAX_METHOD_LEN, "<Object::%s>", cname);
            break;
        default:
            mkey->klass_len = snprintf(result, MAX_METHOD_LEN, "%s", cname);
            break;
    }

    mkey->fqmn_len = snprintf(result, MAX_METHOD_LEN-mkey->klass_len, "%s#%s", result, mname);
    mkey->method_name_len = mkey->fqmn_len - mkey->klass_len;
    mkey->method_name = &result[mkey->klass_len+1]; // skip #

    // append fn_args
    VALUE fn_args = get_fnargs(profile, rb_str_new2(result));
    if (fn_args != Qnil) {
        size_t wtotal = snprintf(result, MAX_METHOD_LEN-mkey->fqmn_len,
            "%s%s", result, StringValuePtr(fn_args));
        mkey->fn_args = &result[mkey->fqmn_len];
        mkey->fn_args_len = wtotal - mkey->fqmn_len;
    }

    return (st_data_t)result;
}

/* ======   Allocation Table  ====== */
st_table *allocations_table_create() { return rb_st_init_numtable(); }

static int allocations_table_free_iterator(st_data_t key, st_data_t value, st_data_t dummy)
{
    prof_allocation_free((prof_allocation_t *)value);
    return ST_CONTINUE;
}

static int prof_method_collect_allocations(st_data_t key, st_data_t value, st_data_t result)
{
    prof_allocation_t *allocation = (prof_allocation_t *)value;
    VALUE arr = (VALUE)result;
    rb_ary_push(arr, prof_allocation_wrap(allocation));
    return ST_CONTINUE;
}

static int prof_method_mark_allocations(st_data_t key, st_data_t value, st_data_t data)
{
    prof_allocation_t *allocation = (prof_allocation_t *)value;
    prof_allocation_mark(allocation);
    return ST_CONTINUE;
}

void allocations_table_free(st_table *table)
{
    rb_st_foreach(table, allocations_table_free_iterator, 0);
    rb_st_free_table(table);
}

/* ================  prof_method_t   =================*/
prof_method_t *prof_get_method(VALUE self)
{
    /* Can't use Data_Get_Struct because that triggers the event hook
       ending up in endless recursion. */
    prof_method_t *result = RTYPEDDATA_DATA(self);

    if (!result)
    {
        rb_raise(rb_eRuntimeError, "This Blackfire::MethodInfo instance has already been freed, "
                                   "likely because its profile has been freed.");
    }

    return result;
}

prof_method_t *prof_method_create(VALUE profile, st_data_t key, method_key_t *mkey, VALUE klass, VALUE msym, VALUE source_file, int source_line)
{
    prof_method_t *result = ALLOC(prof_method_t);

    result->profile = profile;
    result->klass_flags = 0;

    /* Note we do not call resolve_klass_name now because that causes an object
       allocation that shows up in the allocation results so we want to avoid it
       until after the profile run is complete. */

    result->klass = resolve_klass(klass, &result->klass_flags);
    result->klass_name = Qnil;
    result->method_name = msym;
    result->measurement = prof_measurement_create();
    result->mkey = *mkey;
    result->key = key;

    result->call_trees = prof_call_trees_create();
    result->allocations_table = allocations_table_create();

    result->visits = 0;
    result->recursive = false;

    result->object = Qnil;

    result->source_file = source_file;
    result->source_line = source_line;

    result->timespan_selected = -1;

    return result;
}

/* The underlying c structures are freed when the parent profile is freed.
   However, on shutdown the Ruby GC frees objects in any will-nilly order.
   That means the ruby thread object wrapping the c thread struct may
   be freed before the parent profile.  Thus we add in a free function
   for the garbage collector so that if it does get called will nil
   out our Ruby object reference.*/
static void prof_method_ruby_gc_free(void *data)
{
    if (data)
    {
        prof_method_t *method = (prof_method_t *)data;
        method->object = Qnil;
    }
}

static void prof_method_free(prof_method_t *method)
{
    /* Has this method object been accessed by Ruby?  If
       yes clean it up so to avoid a segmentation fault. */
    if (method->object != Qnil)
    {
        RTYPEDDATA(method->object)->data = NULL;
        method->object = Qnil;
    }

    allocations_table_free(method->allocations_table);
    prof_call_trees_free(method->call_trees);
    prof_measurement_free(method->measurement);
    xfree((char *)method->key);
    xfree(method);
}

size_t prof_method_size(const void *data) { return sizeof(prof_method_t); }

void prof_method_mark(void *data)
{
    if (!data)
    {
        return;
    }

    prof_method_t *method = (prof_method_t *)data;

    if (method->profile != Qnil)
    {
        rb_gc_mark(method->profile);
    }

    if (method->object != Qnil)
    {
        rb_gc_mark(method->object);
    }

    rb_gc_mark(method->klass_name);
    rb_gc_mark(method->method_name);
    rb_gc_mark(method->source_file);

    if (method->klass != Qnil)
    {
        rb_gc_mark(method->klass);
    }

    prof_measurement_mark(method->measurement);

    rb_st_foreach(method->allocations_table, prof_method_mark_allocations, 0);
}

static VALUE prof_method_allocate(VALUE klass)
{
    prof_method_t *method_data = prof_method_create(Qnil, Qnil, NULL, Qnil, Qnil, Qnil, 0);
    method_data->object = prof_method_wrap(method_data);
    return method_data->object;
}

static const rb_data_type_t method_info_type = {.wrap_struct_name = "MethodInfo",
                                                .function =
                                                    {
                                                        .dmark = prof_method_mark,
                                                        .dfree = prof_method_ruby_gc_free,
                                                        .dsize = prof_method_size,
                                                    },
                                                .data = NULL,
                                                .flags = RUBY_TYPED_FREE_IMMEDIATELY};

VALUE prof_method_wrap(prof_method_t *method)
{
    if (method->object == Qnil)
    {
        method->object = TypedData_Wrap_Struct(cRpMethodInfo, &method_info_type, method);
    }
    return method->object;
}

st_table *method_table_create() { return st_init_strtable(); }

static int method_table_free_iterator(st_data_t key, st_data_t value, st_data_t dummy)
{
    prof_method_free((prof_method_t *)value);
    return ST_CONTINUE;
}

void method_table_free(st_table *table)
{
    rb_st_foreach(table, method_table_free_iterator, 0);
    rb_st_free_table(table);
}

size_t method_table_insert(st_table *table, st_data_t key, prof_method_t *val) { return rb_st_insert(table, key, (st_data_t)val); }

prof_method_t *method_table_lookup(st_table *table, st_data_t key)
{
    st_data_t val;
    if (rb_st_lookup(table, key, &val))
    {
        return (prof_method_t *)val;
    }

    return NULL;
}

/* ================  Method Info   =================*/
/* Document-class: Blackfire::MethodInfo
   The Blackfire::MethodInfo class stores profiling data for a method.
   One instance of the Blackfire::MethodInfo class is created per method
   called per thread.  Thus, if a method is called in two different
   thread then there will be two Blackfire::MethodInfo objects
   created.  Blackfire::MethodInfo objects can be accessed via
   the Blackfire::Profile object.
*/

/* call-seq:
   allocations -> array

   Returns an array of allocation information.*/
static VALUE prof_method_allocations(VALUE self)
{
    prof_method_t *method = prof_get_method(self);
    VALUE result = rb_ary_new();
    rb_st_foreach(method->allocations_table, prof_method_collect_allocations, result);
    return result;
}

/* call-seq:
   called -> Measurement

   Returns the measurement associated with this method. */
static VALUE prof_method_measurement(VALUE self)
{
    prof_method_t *method = prof_get_method(self);
    return prof_measurement_wrap(method->measurement);
}

/* call-seq:
   source_file => string

   return the source file of the method
*/
static VALUE prof_method_source_file(VALUE self)
{
    prof_method_t *method = prof_get_method(self);
    return method->source_file;
}

/* call-seq:
   line_no -> int

   returns the line number of the method */
static VALUE prof_method_line(VALUE self)
{
    prof_method_t *method = prof_get_method(self);
    return INT2FIX(method->source_line);
}

/* call-seq:
   klass_name -> string

   Returns the name of this method's class.  Singleton classes
   will have the form <Object::Object>. */

static VALUE prof_method_klass_name(VALUE self)
{
    prof_method_t *method = prof_get_method(self);
    if (method->klass_name == Qnil)
    {
        method->klass_name = resolve_klass_name(method->klass, &method->klass_flags);
    }

    return method->klass_name;
}

/* call-seq:
   klass_flags -> integer

   Returns the klass flags */

static VALUE prof_method_klass_flags(VALUE self)
{
    prof_method_t *method = prof_get_method(self);
    return INT2FIX(method->klass_flags);
}

/* call-seq:
   method_name -> string

   Returns the name of this method in the format Object#method.  Singletons
   methods will be returned in the format <Object::Object>#method.*/

static VALUE prof_method_name(VALUE self)
{
    prof_method_t *method = prof_get_method(self);
    return method->method_name;
}

static VALUE prof_method_key(VALUE self)
{
    prof_method_t *method = prof_get_method(self);
    return rb_str_new2((char *)method->key);
}

/* call-seq:
   recursive? -> boolean

   Returns the true if this method is recursively invoked */
static VALUE prof_method_recursive(VALUE self)
{
    prof_method_t *method = prof_get_method(self);
    return method->recursive ? Qtrue : Qfalse;
}

/* call-seq:
   call_trees -> CallTrees

   Returns the CallTrees associated with this method. */
static VALUE prof_method_call_trees(VALUE self)
{
    prof_method_t *method = prof_get_method(self);
    return prof_call_trees_wrap(method->call_trees);
}

/* :nodoc: */
static VALUE prof_method_dump(VALUE self)
{
    prof_method_t *method_data = prof_get_method(self);
    VALUE result = rb_hash_new();

    rb_hash_aset(result, ID2SYM(rb_intern("klass_name")), prof_method_klass_name(self));
    rb_hash_aset(result, ID2SYM(rb_intern("klass_flags")), INT2FIX(method_data->klass_flags));
    rb_hash_aset(result, ID2SYM(rb_intern("method_name")), method_data->method_name);

    rb_hash_aset(result, ID2SYM(rb_intern("key")), INT2FIX(method_data->key));
    rb_hash_aset(result, ID2SYM(rb_intern("recursive")), prof_method_recursive(self));
    rb_hash_aset(result, ID2SYM(rb_intern("source_file")), method_data->source_file);
    rb_hash_aset(result, ID2SYM(rb_intern("source_line")), INT2FIX(method_data->source_line));

    rb_hash_aset(result, ID2SYM(rb_intern("call_trees")), prof_call_trees_wrap(method_data->call_trees));
    rb_hash_aset(result, ID2SYM(rb_intern("measurement")), prof_measurement_wrap(method_data->measurement));
    rb_hash_aset(result, ID2SYM(rb_intern("allocations")), prof_method_allocations(self));

    return result;
}

/* :nodoc: */
static VALUE prof_method_load(VALUE self, VALUE data)
{
    prof_method_t *method_data = prof_get_method(self);
    method_data->object = self;

    method_data->klass_name = rb_hash_aref(data, ID2SYM(rb_intern("klass_name")));
    method_data->klass_flags = FIX2INT(rb_hash_aref(data, ID2SYM(rb_intern("klass_flags"))));
    method_data->method_name = rb_hash_aref(data, ID2SYM(rb_intern("method_name")));
    method_data->key = FIX2LONG(rb_hash_aref(data, ID2SYM(rb_intern("key"))));

    method_data->recursive = rb_hash_aref(data, ID2SYM(rb_intern("recursive"))) == Qtrue ? true : false;

    method_data->source_file = rb_hash_aref(data, ID2SYM(rb_intern("source_file")));
    method_data->source_line = FIX2INT(rb_hash_aref(data, ID2SYM(rb_intern("source_line"))));

    VALUE call_trees = rb_hash_aref(data, ID2SYM(rb_intern("call_trees")));
    method_data->call_trees = prof_get_call_trees(call_trees);

    VALUE measurement = rb_hash_aref(data, ID2SYM(rb_intern("measurement")));
    method_data->measurement = prof_get_measurement(measurement);

    VALUE allocations = rb_hash_aref(data, ID2SYM(rb_intern("allocations")));
    for (int i = 0; i < rb_array_len(allocations); i++)
    {
        VALUE allocation = rb_ary_entry(allocations, i);
        prof_allocation_t *allocation_data = prof_allocation_get(allocation);

        rb_st_insert(method_data->allocations_table, allocation_data->key, (st_data_t)allocation_data);
    }
    return data;
}

void rp_init_method_info()
{
    /* MethodInfo */
    cRpMethodInfo = rb_define_class_under(mProf, "MethodInfo", rb_cObject);
    rb_undef_method(CLASS_OF(cRpMethodInfo), "new");
    rb_define_alloc_func(cRpMethodInfo, prof_method_allocate);

    rb_define_method(cRpMethodInfo, "klass_name", prof_method_klass_name, 0);
    rb_define_method(cRpMethodInfo, "klass_flags", prof_method_klass_flags, 0);
    rb_define_method(cRpMethodInfo, "method_name", prof_method_name, 0);
    rb_define_method(cRpMethodInfo, "key", prof_method_key, 0);

    rb_define_method(cRpMethodInfo, "call_trees", prof_method_call_trees, 0);

    rb_define_method(cRpMethodInfo, "allocations", prof_method_allocations, 0);
    rb_define_method(cRpMethodInfo, "measurement", prof_method_measurement, 0);

    rb_define_method(cRpMethodInfo, "source_file", prof_method_source_file, 0);
    rb_define_method(cRpMethodInfo, "line", prof_method_line, 0);

    rb_define_method(cRpMethodInfo, "recursive?", prof_method_recursive, 0);

    rb_define_method(cRpMethodInfo, "_dump_data", prof_method_dump, 0);
    rb_define_method(cRpMethodInfo, "_load_data", prof_method_load, 1);
}
