require "inifile"
require "os"

module Blackfire
  module Config
    class << self
      attr_writer :local_client_id, :local_client_token, :local_endpoint, :local_server_id,
                  :local_server_token, :local_agent_timeout, :local_log_level, :local_log_file,
                  :local_agent_socket

      def initialize
        @bfconfig = nil
        # this must be initilialized before others
        ini_path = config_file_path
        @bfconfig = IniFile.load(ini_path) unless ini_path.nil?

        @local_client_id = get_data("client-id").to_s
        @local_client_token = get_data("client-token").to_s
        @local_endpoint = get_data("endpoint").to_s
        @local_server_id = get_data("server-id").to_s
        @local_server_token = get_data("server-token").to_s
        @local_agent_timeout = get_data("agent-timeout").gsub("s", "").to_f
        @local_log_level = log_level_to_sym(get_data("log-level"))
        @local_log_file = get_data("log-file").to_s
        @local_agent_socket = get_data("agent-socket").to_s
      end

      def valid_auth_header?
        return false if server_id.nil?
        return false if server_token.nil?
        return false if Query.agent_ids.nil?
        return false if server_id.empty?
        return false if server_token.empty?
        return true if Query.agent_ids.include? server_id

        false
      end

      def client_id
        ret = ENV["BLACKFIRE_CLIENT_ID"]
        return ret if value_is_valid?(ret)

        @local_client_id
      end

      def client_token
        ret = ENV["BLACKFIRE_CLIENT_TOKEN"]
        return ret if value_is_valid?(ret)

        @local_client_token
      end

      def agent_socket
        ret = ENV["BLACKFIRE_AGENT_SOCKET"]
        return ret if value_is_valid?(ret)

        @local_agent_socket
      end

      def endpoint
        ret = ENV["BLACKFIRE_ENDPOINT"]
        return ret if value_is_valid?(ret)

        @local_endpoint
      end

      def server_id
        ret = ENV["BLACKFIRE_SERVER_ID"]
        return ret if value_is_valid?(ret)

        @local_server_id
      end

      def server_token
        ret = ENV["BLACKFIRE_SERVER_TOKEN"]
        return ret if value_is_valid?(ret)

        @local_server_token
      end

      def agent_timeout
        ret = ENV["BLACKFIRE_AGENT_TIMEOUT"]
        return ret.to_f if value_is_valid?(ret)

        @local_agent_timeout
      end

      def log_level
        ret = ENV["BLACKFIRE_LOG_LEVEL"]
        return log_level_to_sym(ret) if value_is_valid?(ret)

        @local_log_level
      end

      def log_file
        ret = ENV["BLACKFIRE_LOG_FILE"]
        return ret if value_is_valid?(ret)

        @local_log_file
      end

      def log_to_file?
        path = get_data("log-file").to_s.downcase
        return false if log_to_stderr?
        return false if log_to_stdout?
        return false if path.empty?

        true
      end

      def log_to_stderr?
        path = get_data("log-file").to_s.downcase
        return true if path == "stderr"
        return true if path.empty?

        false
      end

      def log_to_stdout?
        path = get_data("log-file").to_s.downcase
        return true if path == "stdout"

        false
      end

      def config_file_exists?
        path = config_file_path
        return false if path.nil?

        File.file? path
      end

      def unix_socket?
        agent_socket.include?("unix://")
      end

      private

      def get_data(data_item)
        # First, check if ENV var exists
        env_var_name = "BLACKFIRE_#{data_item.gsub('-', '_').upcase}"
        return ENV[env_var_name] unless ENV[env_var_name].nil?

        # Check if value is defined at ini file
        ret = (@bfconfig["blackfire"][data_item] || "") rescue nil
        return ret if value_is_valid?(ret)

        # last case, use defaults
        defaults(data_item)
      end

      # rubocop:disable Metrics/CyclomaticComplexity
      def defaults(data_item)
        # default values from probe.py
        case data_item
        when "endpoint"
          "https://blackfire.io/"
        when "agent-timeout"
          "0.25"
        when "log-file"
          "stderr"
        when "log-level"
          "1"
        when "agent-socket"
          return "unix:///usr/local/var/run/blackfire-agent.sock" if OS.mac?
          return "socket=tcp://127.0.0.1:8307" if OS.windows?

          "unix:///var/run/blackfire/agent.sock"
        end
      end
      # rubocop:enable Metrics/CyclomaticComplexity

      def config_file_path
        file_name = ".blackfire.ini".freeze

        file_paths = []
        file_paths << build_path_string([ENV["BLACKFIRE_HOME"], file_name])
        file_paths << build_path_string([ENV["XDG_CONFIG_HOME"], file_name]) if OS.linux?
        file_paths << build_path_string([Dir.home, file_name])
        file_paths.each do |fp|
          return fp if File.file? fp
        end
        nil
      end

      def build_path_string(items)
        (items * "/").gsub("//", "/") rescue ""
      end

      def log_level_to_sym(log_level)
        case log_level
        when "1"
          :error
        when "2"
          :warn
        when "3"
          :info
        when "4"
          :debug
        else
          :error
        end
      end

      def value_is_valid?(value)
        if value.kind_of?(String)
          return true unless value.empty?
        else
          return true unless value.nil?
        end

        false
      end
    end
  end
end
