module Blackfire

  class Profiler
    attr_reader :profile, :flag_cpu, :flag_timespan, :timespan_threshold
    attr_accessor :current_connection

    def initialize(query: nil, flag_cpu: nil, flag_timespan: nil, 
                  timespan_threshold: nil, timespan_rules: {}, fnarg_rules: {})
      Query.parse(query)

      # explicit parameters overrides QUERY for better testability
      @flag_cpu = flag_cpu.nil? ? Query.flag_cpu == "1" : flag_cpu
      @flag_timespan = flag_timespan.nil? ? Query.flag_timespan == "1" : flag_timespan
      @timespan_threshold = timespan_threshold.nil? ? Query.timespan_threshold.to_i : timespan_threshold

      @profile = Profile.new(@flag_cpu, @flag_timespan, @timespan_threshold)

      # override timespan/fnarg rules if provided
      timespan_rules.each do | key, value |
        @profile.timespan_rules[key] = value
      end

      fnarg_rules.each do | key, value |
        @profile.fnarg_rules[key] = value
      end

      logger = Logger.build
      logger.info "Profiler initialized. [flag_cpu:#{@flag_cpu}, flag_timespan: #{@flag_timespan}," \
        "timespan_threshold: #{@timespan_threshold}, timespan_rules: #{@profile.timespan_rules}, " \
        "fnarg_rules: #{@profile.fnarg_rules}]"

      @current_connection = nil
    end

    def enable
      # @profile.query = @query # TODO: Move to enable or init or somewhere else?
      @profile.enable
    end

    def disable
      @profile.disable
    end

    def stop
      result = @profile.stop
      logger = Logger.build

      # encode traces
      printer = Blackfire::BlackfireTracesPrinter.new(result)
      result = StringIO.new
      printer.print(result)
      logger.debug result.string
      root_id_line = result.string.split("\n")[0]

      # encode timespan entries
      timespan_entries = ""
      @profile.timespan_entries.each_with_index { |x, i|
        swt, ewt, spr, epr, caller, callee = x

        timespan_entries << "Threshold-#{i}-Start: //#{swt} #{spr}\nThreshold-#{i}-End: #{caller}==>#{callee}//#{ewt} #{epr}\n"
      }

      headers = get_headers + root_id_line
      [headers, timespan_entries, result.string.gsub("#{root_id_line}\n", "")]
    end

    def get_headers
      result = "file-format: BlackfireProbe\nProbe-version: " + Blackfire::VERSION
      result << "\nContext: " + __FILE__.split("/")[-1]
      if Query.profile_title != nil
        result << "\nProfile-Title: " + Query.profile_title
      end
      # TODO: Not sure if dimensions should be like this
      result << "\nCost-Dimensions: wt cpu\nProbed-Os: " + RUBY_PLATFORM
      result << "\nProbed-Runtime: " + RUBY_DESCRIPTION + " - " + RUBY_RELEASE_DATE
      if Query.probed_features != nil
        result << "\nProbed-Language: ruby\nProbed-Features: " + Query.probed_features
      end
      result << "\n"

      result
    end

    def running?
      @profile.running?
    end

    def connected?
      !(@current_connection.nil? || @current_connection.closed?)
    end

    def publish(sentence)
      @current_connection.write sentence
    rescue IOError => e
      logger = Logger.build
      logger.error "Profiler error while trying to publishing: #{e.message}"
      @current_connection.close
    end

    def gets
      begin
        return @current_connection.gets
      rescue IOError => e
        logger = Logger.build
        logger.error "Profiler error while trying to read: #{e.message}"
        @current_connection.close
      end
      nil
    end

    def terminate
      logger = Logger.build
      logger.info "Closing connection"
      @current_connection.close

      nil
    end

    def open_connection
      blackfire_socket = Config.agent_socket
      return if blackfire_socket.empty?

      begin
        unless @current_connection.nil?
          @current_connection.close
          @current_connection = nil
        end

        if Config.unix_socket?
          data = blackfire_socket.split("://")[1]
          @current_connection = UNIXSocket.new(data)
        else
          data = blackfire_socket.split("//")[1].split(":")
          server = data[0]
          port = data[1]
          @current_connection = TCPSocket.new(server, port)
        end
      rescue => e
        logger = Logger.build
        logger.error ([e.message]+e.backtrace).join($/)
        @current_connection = nil
      end
    end

  end
end
