require "stringio"
require "socket"

module Blackfire
  class << self

    def get_timespan_rules(header)
      rules = header.gsub("Blackfire-Timespan: ", "")
      
      rules
    end

    # Starts or resume profiling
    def enable(query = nil)
      logger = Logger.build
      logger.info "Start Blackfire Ruby Probe profiling"

      if defined?(@profiler) && @profiler
        return if @profiler.running?

        @profiler.enable
        return @profiler.profile
      end

      # basic settings
      @profiler = Profiler.new(query: query)
      @profiler.open_connection

      return unless @profiler.connected?

      logger.info "Writing prolog"

      # Blackfire-Query header
      @profiler.publish "Blackfire-Query: #{Query.data}\n"

      # Blackfire-Probe header
      bf_probe_header = "Blackfire-Probe: "

      # "ruby" must be in lowercase to match the language in the metrics
      bf_probe_header << "ruby-#{RUBY_VERSION}-p#{RUBY_PATCHLEVEL}"
      bf_probe_header << ", timespan" if @profiler.flag_timespan
      bf_probe_header << ", blackfire_yml" if BlackfireYAML.defined?
      bf_probe_header << "\n"

      @profiler.publish bf_probe_header

      # Blackfire-Auth header
      @profiler.publish "Blackfire-Auth: #{Config.server_id}:#{Config.server_token}\n" if Config.valid_auth_header?

      # send
      @profiler.publish "\n"

      logger.info "Waiting agent response"
      # wait data
      answer = @profiler.gets

      if answer.start_with? "Blackfire-Response".freeze
        Rack::BlackfireRack.define_header(answer)
        BlackfireYAML.send_file(@profiler) if BlackfireYAML.must_be_sent? answer

        # Check for the Blackfire-Keys header
        header = @profiler.gets
        if header.start_with? "Blackfire-Keys".freeze
          # FIXME: store the keys
          header = @profiler.gets
        end

        loop do
          break if header == "\n" or header == nil

          if header.include? "Blackfire-Timespan"
            timespan_rule = header.gsub("Blackfire-Timespan: ", "")
            @profiler.profile.timespan_rules[timespan_rule[0]][timespan_rule[1..-1].strip!] = true
          end

          if header.include? "Blackfire-Fn-Args"
            fnarg_rule = header.gsub("Blackfire-Fn-Args: ", "")
            fn_name, arg_ids = fnarg_rule.split
            arg_ids = arg_ids.split(',')
            arg_ids.each_with_index { |x, i|
              # convert arg_id to integer if it is a positional argument
              if x.match(/^(\d)+$/) 
                arg_ids[i] = x.to_i
              end
            }
            @profiler.profile.fnarg_rules[fn_name] = arg_ids
          end

          header = @profiler.gets
        end
      else
        logger.error answer
        @profiler = @profiler.terminate
        return
      end

      # start must be last line at the enable
      @profiler.enable

      @profiler.profile
    end

    # Stops profiling
    def end
      logger = Logger.build
      logger.info "Stop called, ending profiling"

      return unless check_running?

      headers, timespan_entries, traces = @profiler.stop

      logger.info "Writing profile"
      data = headers + "\n" + timespan_entries + "\n" + traces
      @profiler.publish data

      # ends profiler
      @profiler = @profiler.terminate

      data
    end

    def check_running?
      return false unless defined?(@profiler) && @profiler

      @profiler.running?
    end
  end
end
